from ..Utils import utils, FetchError, AutoRepr
from cachetools.func import ttl_cache

from ..DataHandler import Endpoint
endpoint = Endpoint()

class OnlinePlayer(AutoRepr):
    def __init__(self, player):
        self.name = player["account"]
        self.world = player["world"]
        self.x = player["x"]
        self.y = player["y"]
        self.z = player["z"]

class Players:
    def __init__(self, map, towns):
        self.Towns = towns

        self.online = self.online(map)
        self.residents = self.residents(self)
        self.townless = self.townless(self)

    @ttl_cache(2, 120)
    def all(self): 
        names = []
        names.extend(self.residents.all())

        allTownless = self.townless.all()
        for pl in allTownless:
            names.append(pl['name'])
            
        return names

    class residents: 
        def __init__(self, players):
            self.Towns = players.Towns

        @ttl_cache(8, 120)
        def all(self):
            output: list[str] = []
            allTowns = self.Towns.all()

            for t in allTowns:
                townResidents: list[str] = t['residents']

                for res in townResidents:
                    output.append(res)

            return output

        @ttl_cache(8, 120)
        def get(self, resName: str, resList: list[str] | None):
            if resList is None:
                resList = self.all()

            res = utils.find(lambda res: res.lower() == resName.lower(), resList)
            if res is None:
                return "Could not find resident '" + resName + "'"

            return res

    class townless:
        def __init__(self, players):
            self.resList = players.residents.all()
            self.ops = players.online.all()

        @ttl_cache(8, 2)    
        def all(self):
            output = []

            for p in self.ops:
                foundRes = utils.find(lambda res: res.lower() == p['name'].lower(), self.resList)
                if foundRes is not None: continue
                
                output.append(p)

            return output

    class online:
        def __init__(self, map):
            self.map = map

        @ttl_cache(8, 2)
        def all(self):
            output = []

            try: playerData = endpoint.fetch('players', self.map)
            except FetchError as e: 
                print(e)
                return []

            pData = playerData["players"]
            for op in pData: 
                output.append(vars(OnlinePlayer(op)))

            return output

        def find(self, playerName): return self.get(playerName)

        @ttl_cache(8, 2)
        def get(self, playerName, ops=None):
            if ops is None: 
                ops = self.all()

            pName = playerName.lower()
            foundPlayer = utils.find(lambda player: player['name'].lower() == pName, ops)
        
            if foundPlayer is None: return "Could not find player '" + playerName + "'" 
            return foundPlayer


from ..Utils import FetchError, utils, AutoRepr
from cachetools.func import ttl_cache
from ..DataHandler import Endpoint
endpoint = Endpoint()

class Town(AutoRepr):
    def __init__(self,
        name="", nation="No Nation",
        mayor="", area=0, x=0, z=0,
        residents=None, flags=None, colourCodes=None
    ):
        if residents is None:
            residents = []

        if flags is None:
            flags = {}

        if colourCodes is None:
            colourCodes = {}

        self.name = name
        self.nation = nation
        self.mayor = mayor
        self.area = area
        self.x = x
        self.z = z
        self.residents = residents
        self.flags = flags
        self.colourCodes = colourCodes

class Towns:
    def __init__(self, map_name):
        self.map_name = map_name
        print("Created new 'Towns' instance.")

    @ttl_cache(16, 120)
    def all(self):
        towns_array = []

        try:
            map_data = endpoint.fetch('map', self.map_name)
        except FetchError as e:
            print(e)
            return []

        markerset = map_data["sets"]['townyPlugin.markerset']
        areas = markerset['areas']
        town_area_names = list(areas.keys())

        for i in range(len(town_area_names)):
            town = areas[town_area_names[i]]
            raw_info = town["desc"].split("<br />")

            info = []

            for x in raw_info:
                info.append(utils.striptags(x))

            if "Shop" in info[0]:
                continue

            mayor = info[1][7:]
            if mayor == "":
                continue

            split = info[0].split(" (")
            nation_name = (split[1] if len(split) < 3 else split[2])[0:-1]
            residents = info[2][9:].split(", ")

            x = round((max(town["x"]) + min(town["x"])) / 2)
            z = round((max(town["z"]) + min(town["z"])) / 2)

            flags = {
                'pvp': utils.strAsBool(info[4][5:]),
                'mobs': utils.strAsBool(info[5][6:]),
                'public': utils.strAsBool(info[6][8:]),
                'explosions': utils.strAsBool(info[7][11:]),
                'fire': utils.strAsBool(info[8][6:]),
                'capital': utils.strAsBool(info[9][9:])
            }

            colourCodes = {
                'fill': town['fillcolor'],
                'outline': town['color']
            }

            ct = Town(
                name=utils.removeStyleCharacters(town['label']),
                mayor=mayor,
                area=utils.townArea(town),
                x=x, z=z,
                residents=residents,
                flags=flags,
                colourCodes=colourCodes
            )

            if nation_name != '':
                ct.nation = nation_name.strip()

            towns_array.append(ct)

        cached_towns: list[dict[str, Town]]  = []
        temp = {}

        for t in towns_array:
            if temp.get(t.name, None) is None:
                temp[t.name] = t
                cached_towns.append(vars(t))
            else:
                temp[t.name].area += t.area

        return cached_towns

    def get(self, town_name, towns=None):
        if towns is None:
            towns = self.all()
        found_town = utils.find(lambda town: town['name'].lower() == town_name.lower(), towns)

        if found_town is None:
            return "Could not find town '" + town_name + "'"
        return found_town

    def nearby(self, x_input, z_input, x_radius, z_radius, towns=None):
        if towns is None:
            towns = self.all()

        filtered_towns = []
        for t in towns:
            distance_x = t['x'] - x_input
            distance_z = t['z'] - z_input

            if -x_radius <= distance_x <= x_radius and -z_radius <= distance_z <= z_radius:
                filtered_towns.append(t)

        return filtered_towns if filtered_towns else []


from enum import Enum
import math

from .Towns import Town
from ..Utils import utils

from typing import TypedDict

import asyncio
from pyee import EventEmitter

LocationType = TypedDict('LocationType', { 'x': int, 'z': int })
class Location:
    def __init__(self, x, z):
        self.x = x
        self.z = z

class RouteOptions:
    avoid_pvp: bool
    avoid_public: bool

    def __init__(self, pvp, public):
        self.avoid_pvp = pvp
        self.avoid_public = public

class Route(Enum):
    SAFEST = RouteOptions(True, True)
    FASTEST = RouteOptions(False, False)
    AVOID_PVP = RouteOptions(True, False)
    AVOID_PUBLIC = RouteOptions(False, True)

class RouteInfo:
    nation: object
    distance: int
    direction: str

    def __init__(self, nation, distance, direction):
        self.nation = nation
        self.distance = distance
        self.direction = direction

class GPS(EventEmitter):
    map: None
    emitted_underground: bool
    last_loc: {}

    def __init__(self, map):
        # Initialize event emitter
        super().__init__()

        # Used when `track()` is called.
        self.emitted_underground = False
        self.last_loc = None

        # The parent Map the GPS was set up on.
        self.map = map

    async def track(self, player_name: str, interval=3000, route: RouteOptions = Route.FASTEST.value):
        async def track_interval():
            while True:
                player = await self.map.Players.get(player_name)
                playerX, playerZ, playerWorld = player['x'], player['z'], player['world']

                if not playerWorld:
                    self.emit("error", {
                        "err": "INVALID_PLAYER",
                        "msg": "Player is offline or does not exist!"
                    })

                    return

                underground = playerX == 0 and playerZ == 0 and playerWorld != "some-other-bogus-world"
                if underground:
                    if not self.emitted_underground:
                        self.emitted_underground = True

                        if not self.last_loc:
                            self.emit("underground", "No last location. Waiting for this player to show.")
                            return

                        try:
                            route_info = self.find_route(self.last_loc, route)
                            self.emit("underground", {
                                "lastLocation": self.last_loc,
                                "routeInfo": route_info
                            })
                        except Exception as e:
                            self.emit("error", {
                                "err": "INVALID_LAST_LOC",
                                "msg": str(e)
                            })
                else:
                    self.last_loc = {
                        "x": playerX,
                        "z": playerZ
                    }

                    try:
                        route_info = self.find_route(self.last_loc, route)
                        self.emit("locationUpdate", route_info)
                    except Exception as e:
                        self.emit("error", {"err": "INVALID_LOC", "msg": str(e)})

                await asyncio.sleep(interval / 1000)

        asyncio.create_task(track_interval())

    def find_safest_route(self, loc: LocationType):
        return self.find_route(loc, Route.SAFEST.value)

    def find_fastest_route(self, loc: LocationType):
        return self.find_route(loc, Route.FASTEST.value)

    def find_route(self, loc: LocationType, route: RouteOptions):
        towns = self.map.Towns.all()
        nations = self.map.Nations.all()

        filtered = []
        for n in nations:
            capital: Town = next((t for t in towns if t.name == n['capital']['name']), None)
            if capital is None: continue

            PVP = route.avoid_pvp and capital.flags['pvp']
            PUBLIC = route.avoid_public and (capital.flags['public'] is None)
            if PVP or PUBLIC: continue

            filtered.append(n)

        if len(filtered) < 1:
            return None

        distance, nation = 0.0, None
        for n in filtered:
            closest = GPS.find_closest(distance, n, loc)
            distance, nation = closest['distance'], closest['nation']

        direction = GPS.cardinal_direction(nation['capital'], loc)
        return RouteInfo(nation, distance, direction)

    @staticmethod
    def find_closest(dist, nation, loc: LocationType):
        capital = nation["capital"]
        curDist = utils.manhattan_distance(capital, loc)

        if dist < curDist:
            return dist

        return {
            "distance": round(curDist),
            "nation": {
                "name": nation["name"],
                "capital": capital
            }
        }

    @staticmethod
    def cardinal_direction(loc1: LocationType, loc2: LocationType):
        delta_x = loc2['x'] - loc1['x']
        delta_z = loc2['z'] - loc1['z']

        angle_rad = math.atan2(delta_z, delta_x)
        angle_deg = (angle_rad * 180) / math.pi

        if -45 <= angle_deg < 45:
            return "east"
        elif 45 <= angle_deg < 135:
            return "north"
        elif angle_deg >= 135 or angle_deg < -135:
            return "west"
        else:
            return "south"



from .Towns import Towns
from cachetools.func import ttl_cache
from ..Utils import utils, AutoRepr

class Nation(AutoRepr):
    def __init__(self, name="", king="", capital=None, residents=None, towns=None, area=0):
        if capital is None: capital = {}
        if towns is None: towns = []
        if residents is None: residents = []

        self.name = name
        self.king = king
        self.capital = capital
        self.residents = residents
        self.towns = towns
        self.area = area

    #def __sub__(self, key, other):
        #return getattr(self, key) - other

class Nations:
    def __init__(self, mapName):
        self.Towns = Towns(mapName)
        self.Nations = self

        print("Created new 'nations' instance")

    @ttl_cache(4, 120)
    def get(self, nationName, nations=None):
        if nations is None:
            nations = self.all()

        foundNation = utils.find(lambda n: n['name'].lower() == nationName.lower(), nations)
        if foundNation is None:
            return "Could not find nation '" + nationName + "'"

        return foundNation

    @ttl_cache(16, 120)
    def all(self):
        raw = {}
        output: list[dict[str, Nation]] = []

        for town in self.Towns.all():
            nationName = town["nation"]
            if nationName == 'No Nation':
                continue

            # Doesn't already exist, create new nation.
            if raw.get(nationName, None) is None:
                raw[nationName] = Nation(
                    name=nationName,
                    residents=town['residents'],
                    towns=[],
                    area=0
                )

                output.append(vars(raw[nationName]))

            # Add up existing values
            townName = town['name']
            raw[nationName].area += town['area']

            raw[nationName].residents.extend(town['residents'])
            raw[nationName].residents = utils.listFromDictKey(raw[nationName].residents)

            if raw[nationName].name == town['nation']:
                raw[nationName].towns.append(townName)

            if town['flags']['capital'] is True:
                raw[nationName].king = town['mayor']
                raw[nationName].capital = {
                    'name': townName,
                    'x': town['x'],
                    'z': town['z']
                }

        return output

    def nearby(self, x_input, z_input, x_radius, z_radius, nations=None):
        if nations is None:
            nations = self.all()

        filtered_nations = []
        for n in nations:
            distance_x = n['x'] - x_input
            distance_z = n['z'] - z_input

            if -x_radius <= distance_x <= x_radius and -z_radius <= distance_z <= z_radius:
                filtered_nations.append(n)

        return filtered_nations if filtered_nations else []
